from collections.abc import Mapping
from typing import Any, TypeVar, TYPE_CHECKING

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

from dateutil.parser import isoparse
from typing import cast
from typing import Union
import datetime

if TYPE_CHECKING:
    from ..models.flag_value import FlagValue
    from ..models.flag_environments import FlagEnvironments


T = TypeVar("T", bound="Flag")


@_attrs_define
class Flag:
    """
    Example:
        {'created_at': '2026-03-27T10:00:00Z', 'default': False, 'description': 'Enable dark mode for the application
            UI', 'environments': {'production': {'default': False, 'enabled': True, 'rules': [{'description': 'Beta users
            get dark mode', 'logic': {'attribute': 'beta', 'op': 'eq', 'value': True}, 'value': True}]}, 'staging':
            {'default': True, 'enabled': True, 'rules': []}}, 'key': 'dark_mode', 'name': 'Dark Mode', 'type': 'BOOLEAN',
            'updated_at': '2026-03-27T10:00:00Z', 'values': [{'name': 'on', 'value': True}, {'name': 'off', 'value':
            False}]}

    Attributes:
        key (str): Unique key within account
        name (str): Human-readable display name
        type_ (str): Value type: STRING, BOOLEAN, NUMERIC, or JSON
        default (Any): Default value; must reference a value in the values array
        values (list['FlagValue']): Closed set of possible values
        description (Union[Unset, str]):  Default: ''.
        environments (Union[Unset, FlagEnvironments]):
        created_at (Union[None, Unset, datetime.datetime]):
        updated_at (Union[None, Unset, datetime.datetime]):
    """

    key: str
    name: str
    type_: str
    default: Any
    values: list["FlagValue"]
    description: Union[Unset, str] = ""
    environments: Union[Unset, "FlagEnvironments"] = UNSET
    created_at: Union[None, Unset, datetime.datetime] = UNSET
    updated_at: Union[None, Unset, datetime.datetime] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        key = self.key

        name = self.name

        type_ = self.type_

        default = self.default

        values = []
        for values_item_data in self.values:
            values_item = values_item_data.to_dict()
            values.append(values_item)

        description = self.description

        environments: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.environments, Unset):
            environments = self.environments.to_dict()

        created_at: Union[None, Unset, str]
        if isinstance(self.created_at, Unset):
            created_at = UNSET
        elif isinstance(self.created_at, datetime.datetime):
            created_at = self.created_at.isoformat()
        else:
            created_at = self.created_at

        updated_at: Union[None, Unset, str]
        if isinstance(self.updated_at, Unset):
            updated_at = UNSET
        elif isinstance(self.updated_at, datetime.datetime):
            updated_at = self.updated_at.isoformat()
        else:
            updated_at = self.updated_at

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "key": key,
                "name": name,
                "type": type_,
                "default": default,
                "values": values,
            }
        )
        if description is not UNSET:
            field_dict["description"] = description
        if environments is not UNSET:
            field_dict["environments"] = environments
        if created_at is not UNSET:
            field_dict["created_at"] = created_at
        if updated_at is not UNSET:
            field_dict["updated_at"] = updated_at

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.flag_value import FlagValue
        from ..models.flag_environments import FlagEnvironments

        d = dict(src_dict)
        key = d.pop("key")

        name = d.pop("name")

        type_ = d.pop("type")

        default = d.pop("default")

        values = []
        _values = d.pop("values")
        for values_item_data in _values:
            values_item = FlagValue.from_dict(values_item_data)

            values.append(values_item)

        description = d.pop("description", UNSET)

        _environments = d.pop("environments", UNSET)
        environments: Union[Unset, FlagEnvironments]
        if isinstance(_environments, Unset):
            environments = UNSET
        else:
            environments = FlagEnvironments.from_dict(_environments)

        def _parse_created_at(data: object) -> Union[None, Unset, datetime.datetime]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, str):
                    raise TypeError()
                created_at_type_0 = isoparse(data)

                return created_at_type_0
            except:  # noqa: E722
                pass
            return cast(Union[None, Unset, datetime.datetime], data)

        created_at = _parse_created_at(d.pop("created_at", UNSET))

        def _parse_updated_at(data: object) -> Union[None, Unset, datetime.datetime]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, str):
                    raise TypeError()
                updated_at_type_0 = isoparse(data)

                return updated_at_type_0
            except:  # noqa: E722
                pass
            return cast(Union[None, Unset, datetime.datetime], data)

        updated_at = _parse_updated_at(d.pop("updated_at", UNSET))

        flag = cls(
            key=key,
            name=name,
            type_=type_,
            default=default,
            values=values,
            description=description,
            environments=environments,
            created_at=created_at,
            updated_at=updated_at,
        )

        flag.additional_properties = d
        return flag

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



"""
Smpl Flags SDK Showcase — Management API
==========================================

Demonstrates the smplkit Python SDK's management plane for Smpl Flags:

- Client initialization
- Creating flags (BOOLEAN, STRING, NUMERIC, JSON) via FlagType enum
- Rule builder: fluent API for constructing JSON Logic rules
- Configuring values, environments, and rules
- Convenience methods (addRule)
- Updating flag definitions
- Listing and inspecting flags
- Deleting flags
- Managing context types

Most customers will create and configure flags via the Console UI.
This showcase demonstrates the programmatic equivalent — useful for
infrastructure-as-code, CI/CD pipelines, setup scripts, and automated
testing.

For the runtime evaluation experience (declaring flags in code,
evaluating them, context providers, caching, live updates), see
``flags_runtime_showcase.py``.

Prerequisites:
    - ``pip install smplkit-sdk``
    - A valid smplkit API key (set via ``SMPLKIT_API_KEY`` env var)
    - The smplkit Flags service running and reachable
    - At least two environments configured (e.g., ``staging``, ``production``)

Usage::

    export SMPLKIT_API_KEY="sk_api_..."
    python examples/flags_management_showcase.py
"""

import asyncio
import os
import sys

from smplkit import AsyncSmplClient, FlagType, Rule

# ---------------------------------------------------------------------------
# Configuration
# ---------------------------------------------------------------------------

API_KEY = os.environ.get("SMPLKIT_API_KEY", "")

if not API_KEY:
    print("ERROR: Set the SMPLKIT_API_KEY environment variable before running.")
    print("  export SMPLKIT_API_KEY='sk_api_...'")
    sys.exit(1)

# ---------------------------------------------------------------------------
# Helpers
# ---------------------------------------------------------------------------


def section(title: str) -> None:
    """Print a section header for readability."""
    print(f"\n{'=' * 60}")
    print(f"  {title}")
    print(f"{'=' * 60}\n")


def step(description: str) -> None:
    """Print a step within a section."""
    print(f"  → {description}")


async def main() -> None:

    # ======================================================================
    # 1. SDK INITIALIZATION
    # ======================================================================
    section("1. SDK Initialization")

    client = AsyncSmplClient(API_KEY)
    step("AsyncSmplClient initialized")

    # ======================================================================
    # 2. CREATE FLAGS
    # ======================================================================
    #
    # Flags are created with a key, name, type, default value, and an
    # optional values array. The key is the lone positional arg; everything
    # else is a keyword arg.
    #
    # Required: key (positional), name, type (FlagType enum), default
    # Optional: description, values
    #
    # FlagType enum: FlagType.BOOLEAN, FlagType.STRING, FlagType.NUMERIC,
    #                FlagType.JSON
    #
    # For BOOLEAN flags, the values array is auto-generated if not
    # provided: [{"name": "True", "value": true}, {"name": "False", "value": false}].
    #
    # For STRING, NUMERIC, and JSON flags, the values array defines the
    # closed set of legal values the flag can serve. The default must
    # reference a value in this set.
    # ======================================================================

    # ------------------------------------------------------------------
    # 2a. BOOLEAN flag
    # ------------------------------------------------------------------
    section("2a. Create a Boolean Flag")

    checkout_flag = await client.flags.create(
        "checkout-v2",
        name="Checkout V2",
        type=FlagType.BOOLEAN,
        default=False,
        description="Controls rollout of the new checkout experience.",
    )
    step(f"Created: key={checkout_flag.key}, type={checkout_flag.type}")
    step(f"  id={checkout_flag.id}")
    step(f"  values={checkout_flag.values}")
    step(f"  default={checkout_flag.default}")
    # values auto-generated: [{"name": "True", "value": true}, {"name": "False", "value": false}]

    # ------------------------------------------------------------------
    # 2b. STRING flag
    # ------------------------------------------------------------------
    section("2b. Create a String Flag")

    banner_flag = await client.flags.create(
        "banner-color",
        name="Banner Color",
        type=FlagType.STRING,
        default="red",
        description="Controls the banner color shown to users.",
        values=[
            {"name": "Red", "value": "red"},
            {"name": "Green", "value": "green"},
            {"name": "Blue", "value": "blue"},
        ],
    )
    step(f"Created: key={banner_flag.key}, type={banner_flag.type}")
    step(f"  values={banner_flag.values}")

    # ------------------------------------------------------------------
    # 2c. NUMERIC flag
    # ------------------------------------------------------------------
    section("2c. Create a Numeric Flag")

    retry_flag = await client.flags.create(
        "max-retries",
        name="Max Retries",
        type=FlagType.NUMERIC,
        default=3,
        description="Maximum number of API retries before failing.",
        values=[
            {"name": "Low (1)", "value": 1},
            {"name": "Standard (3)", "value": 3},
            {"name": "High (5)", "value": 5},
            {"name": "Aggressive (10)", "value": 10},
        ],
    )
    step(f"Created: key={retry_flag.key}, type={retry_flag.type}")

    # ------------------------------------------------------------------
    # 2d. JSON flag
    # ------------------------------------------------------------------
    section("2d. Create a JSON Flag")

    theme_flag = await client.flags.create(
        "ui-theme",
        name="UI Theme",
        type=FlagType.JSON,
        default={"mode": "light", "accent": "#0066cc"},
        description="Controls the UI theme configuration.",
        values=[
            {"name": "Light", "value": {"mode": "light", "accent": "#0066cc"}},
            {"name": "Dark", "value": {"mode": "dark", "accent": "#66ccff"}},
            {"name": "High Contrast", "value": {"mode": "dark", "accent": "#ffffff"}},
        ],
    )
    step(f"Created: key={theme_flag.key}, type={theme_flag.type}")

    # ======================================================================
    # 3. CONFIGURE ENVIRONMENTS AND RULES
    # ======================================================================
    #
    # Each flag can be independently configured per environment. An
    # environment entry includes:
    #   - enabled (bool): whether rules are evaluated
    #   - default (optional): environment-level override of the flag default
    #   - rules (array): ordered list of rules (first match wins)
    #
    # Rules can be built using the Rule builder (recommended) or as raw
    # JSON Logic dicts. The Rule builder provides a fluent API:
    #
    #   Rule("description")
    #       .when("user.plan", "==", "enterprise")
    #       .serve(True)
    #       .build()
    #
    # Multiple .when() calls are AND'd. .serve() sets the value.
    # .build() finalizes the rule as a dict ready for the API.
    # .environment() is optional — used with addRule (see section 5).
    #
    # Supported operators: ==, !=, >, <, >=, <=, in, contains
    # ======================================================================

    # ------------------------------------------------------------------
    # 3a. Configure checkout-v2 environments
    # ------------------------------------------------------------------
    section("3a. Configure checkout-v2 Environments")

    await checkout_flag.update(
        environments={
            "staging": {
                "enabled": True,
                "rules": [
                    Rule("Enable for enterprise users in US region")
                        .when("user.plan", "==", "enterprise")
                        .when("account.region", "==", "us")
                        .serve(True)
                        .build(),
                    Rule("Enable for beta testers")
                        .when("user.beta_tester", "==", True)
                        .serve(True)
                        .build(),
                ],
            },
            "production": {
                "enabled": False,
                "default": False,
                "rules": [],
            },
        },
    )
    step("staging: enabled with 2 targeting rules")
    step("production: disabled, default=false")

    # ------------------------------------------------------------------
    # 3b. Configure banner-color environments
    # ------------------------------------------------------------------
    section("3b. Configure banner-color Environments")

    await banner_flag.update(
        environments={
            "staging": {
                "enabled": True,
                "rules": [
                    Rule("Blue for enterprise users")
                        .when("user.plan", "==", "enterprise")
                        .serve("blue")
                        .build(),
                    Rule("Green for technology companies")
                        .when("account.industry", "==", "technology")
                        .serve("green")
                        .build(),
                ],
            },
            "production": {
                "enabled": True,
                "default": "blue",
                "rules": [],
            },
        },
    )
    step("staging: enabled with 2 rules")
    step("production: enabled, no rules, default override = blue")

    # ------------------------------------------------------------------
    # 3c. Configure max-retries environments
    # ------------------------------------------------------------------
    section("3c. Configure max-retries Environments")

    await retry_flag.update(
        environments={
            "staging": {
                "enabled": True,
                "rules": [
                    Rule("High retries for large accounts")
                        .when("account.employee_count", ">", 100)
                        .serve(5)
                        .build(),
                ],
            },
            "production": {
                "enabled": True,
                "rules": [],
            },
        },
    )
    step("staging: enabled with 1 rule")
    step("production: enabled, no rules")

    # ======================================================================
    # 4. INSPECT AND LIST FLAGS
    # ======================================================================

    section("4. List and Inspect Flags")

    # List all flags.
    flags = await client.flags.list()
    step(f"Total flags: {len(flags)}")
    for f in flags:
        env_keys = list(f.environments.keys()) if f.environments else []
        step(f"  {f.key} ({f.type}) — default={f.default}, environments={env_keys}")

    # Fetch a single flag by ID.
    fetched = await client.flags.get(checkout_flag.id)
    step(f"\nFetched by ID: {fetched.key}")
    step(f"  staging rules: {len(fetched.environments.get('staging', {}).get('rules', []))}")
    step(f"  production enabled: {fetched.environments.get('production', {}).get('enabled')}")

    # ======================================================================
    # 5. UPDATE A FLAG
    # ======================================================================

    section("5. Update a Flag")

    # Add a new value to banner-color.
    step("Adding 'Purple' to banner-color values...")
    current_values = banner_flag.values
    current_values.append({"name": "Purple", "value": "purple"})
    await banner_flag.update(values=current_values)
    step(f"Updated values: {[v['name'] for v in banner_flag.values]}")

    # Change the flag-level default.
    step("Changing banner-color default to 'blue'...")
    await banner_flag.update(default="blue")
    step(f"Updated default: {banner_flag.default}")

    # Add a rule to an existing environment — the hard way (raw JSON Logic).
    # You have to fetch current state, build the dict, append, and send
    # the whole environments object.
    step("Adding a rule to banner-color production (raw JSON Logic)...")
    current_envs = banner_flag.environments
    prod = current_envs.get("production", {"enabled": True, "rules": []})
    prod_rules = prod.get("rules", [])
    prod_rules.append({
        "description": "Purple for enterprise users",
        "logic": {"==": [{"var": "user.plan"}, "enterprise"]},
        "value": "purple",
    })
    await banner_flag.update(
        environments={
            **current_envs,
            "production": {**prod, "rules": prod_rules},
        },
    )
    step(f"Production now has {len(prod_rules)} rule(s)")

    # Add a rule — the easy way. addRule takes a single built Rule.
    # The Rule's .environment() tells addRule where to insert it.
    step("Adding another rule to banner-color production (addRule + Rule)...")
    await banner_flag.addRule(
        Rule("Green for retail companies")
            .environment("production")
            .when("account.industry", "==", "retail")
            .serve("green")
            .build()
    )
    step("Rule added — no manual environment juggling, no raw JSON Logic")

    # Verify both rules are there.
    refreshed = await client.flags.get(banner_flag.id)
    prod_rules = refreshed.environments.get("production", {}).get("rules", [])
    step(f"Production rules after both additions: {len(prod_rules)}")
    for i, rule in enumerate(prod_rules):
        step(f"  [{i}] {rule.get('description', 'no description')}")

    # ======================================================================
    # 6. CONTEXT TYPE MANAGEMENT
    # ======================================================================
    #
    # Context types define the shape of data that rules can target. They
    # are typically auto-created by the SDK during runtime context
    # registration, but can also be managed explicitly via the API —
    # useful for setting up the Console rule builder before any SDK is
    # deployed.
    # ======================================================================

    section("6. Context Type Management")

    # Create context types that the Console rule builder will use.
    user_ct = await client.flags.create_context_type(
        "user",
        name="User",
    )
    step(f"Created context type: key={user_ct.key}, name={user_ct.name}")

    # Add known attributes.
    await client.flags.update_context_type(
        user_ct.id,
        attributes={
            "first_name": {},
            "plan": {},
            "beta_tester": {},
        },
    )
    step(f"Added attributes: {list(user_ct.attributes.keys())}")

    account_ct = await client.flags.create_context_type(
        "account",
        name="Account",
    )
    await client.flags.update_context_type(
        account_ct.id,
        attributes={
            "industry": {},
            "region": {},
            "employee_count": {},
        },
    )
    step(f"Created context type: key={account_ct.key}")

    # List context types.
    context_types = await client.flags.list_context_types()
    for ct in context_types:
        attrs = list(ct.attributes.keys()) if ct.attributes else []
        step(f"  {ct.key}: attributes={attrs}")

    # ======================================================================
    # 7. CLEANUP
    # ======================================================================
    section("7. Cleanup")

    # Delete flags.
    for flag in [checkout_flag, banner_flag, retry_flag, theme_flag]:
        await client.flags.delete(flag.id)
        step(f"Deleted flag: {flag.key}")

    # Delete context types.
    for ct in [user_ct, account_ct]:
        await client.flags.delete_context_type(ct.id)
        step(f"Deleted context type: {ct.key}")

    # Close the client.
    await client.close()
    step("AsyncSmplClient closed")

    # ======================================================================
    # DONE
    # ======================================================================
    section("ALL DONE")
    print("  The Flags Management showcase completed successfully.")
    print("  All flags and context types have been cleaned up.\n")


if __name__ == "__main__":
    asyncio.run(main())



"""Tests for context conversion and local JSON Logic evaluation."""

from smplkit.flags.client import _contexts_to_eval_dict, _evaluate_flag, _hash_context
from smplkit.flags.types import Context


class TestContextsToEvalDict:
    def test_basic_conversion(self):
        contexts = [
            Context("user", "u-123", plan="enterprise"),
            Context("account", "acme", region="us"),
        ]
        result = _contexts_to_eval_dict(contexts)
        assert result == {
            "user": {"key": "u-123", "plan": "enterprise"},
            "account": {"key": "acme", "region": "us"},
        }

    def test_key_injected(self):
        contexts = [Context("device", "d-1", os="ios")]
        result = _contexts_to_eval_dict(contexts)
        assert result["device"]["key"] == "d-1"
        assert result["device"]["os"] == "ios"

    def test_empty_list(self):
        assert _contexts_to_eval_dict([]) == {}

    def test_no_attributes(self):
        result = _contexts_to_eval_dict([Context("user", "u-1")])
        assert result == {"user": {"key": "u-1"}}


class TestHashContext:
    def test_same_input_same_hash(self):
        d1 = {"user": {"key": "u-1", "plan": "enterprise"}}
        d2 = {"user": {"plan": "enterprise", "key": "u-1"}}  # different order
        assert _hash_context(d1) == _hash_context(d2)

    def test_different_input_different_hash(self):
        d1 = {"user": {"key": "u-1"}}
        d2 = {"user": {"key": "u-2"}}
        assert _hash_context(d1) != _hash_context(d2)


class TestEvaluateFlag:
    """Test _evaluate_flag following ADR-022 §2.6 semantics."""

    def _make_flag(self, *, default, environments=None):
        return {
            "key": "test-flag",
            "name": "Test",
            "type": "BOOLEAN",
            "default": default,
            "values": [],
            "environments": environments or {},
        }

    def test_no_environment_returns_flag_default(self):
        flag = self._make_flag(default=False)
        assert _evaluate_flag(flag, "staging", {}) is False

    def test_environment_not_found_returns_flag_default(self):
        flag = self._make_flag(default="red", environments={"production": {"enabled": True, "rules": []}})
        assert _evaluate_flag(flag, "staging", {}) == "red"

    def test_disabled_returns_env_default(self):
        flag = self._make_flag(
            default=False,
            environments={"staging": {"enabled": False, "default": True, "rules": []}},
        )
        assert _evaluate_flag(flag, "staging", {}) is True

    def test_disabled_no_env_default_returns_flag_default(self):
        flag = self._make_flag(
            default="fallback",
            environments={"staging": {"enabled": False, "rules": []}},
        )
        assert _evaluate_flag(flag, "staging", {}) == "fallback"

    def test_first_matching_rule_wins(self):
        flag = self._make_flag(
            default=False,
            environments={
                "staging": {
                    "enabled": True,
                    "rules": [
                        {"logic": {"==": [{"var": "user.plan"}, "enterprise"]}, "value": True},
                        {"logic": {"==": [{"var": "user.plan"}, "enterprise"]}, "value": False},
                    ],
                }
            },
        )
        ctx = {"user": {"plan": "enterprise"}}
        assert _evaluate_flag(flag, "staging", ctx) is True

    def test_no_rules_match_returns_env_default(self):
        flag = self._make_flag(
            default="flag-default",
            environments={
                "staging": {
                    "enabled": True,
                    "default": "env-default",
                    "rules": [
                        {"logic": {"==": [{"var": "user.plan"}, "enterprise"]}, "value": "matched"},
                    ],
                }
            },
        )
        ctx = {"user": {"plan": "free"}}
        assert _evaluate_flag(flag, "staging", ctx) == "env-default"

    def test_no_rules_match_no_env_default_returns_flag_default(self):
        flag = self._make_flag(
            default="flag-default",
            environments={
                "staging": {
                    "enabled": True,
                    "rules": [
                        {"logic": {"==": [{"var": "user.plan"}, "enterprise"]}, "value": "matched"},
                    ],
                }
            },
        )
        ctx = {"user": {"plan": "free"}}
        assert _evaluate_flag(flag, "staging", ctx) == "flag-default"

    def test_numeric_comparison(self):
        flag = self._make_flag(
            default=3,
            environments={
                "staging": {
                    "enabled": True,
                    "rules": [
                        {"logic": {">": [{"var": "account.employee_count"}, 100]}, "value": 5},
                    ],
                }
            },
        )
        ctx = {"account": {"employee_count": 500}}
        assert _evaluate_flag(flag, "staging", ctx) == 5

    def test_and_condition(self):
        flag = self._make_flag(
            default=False,
            environments={
                "staging": {
                    "enabled": True,
                    "rules": [
                        {
                            "logic": {
                                "and": [
                                    {"==": [{"var": "user.plan"}, "enterprise"]},
                                    {"==": [{"var": "account.region"}, "us"]},
                                ]
                            },
                            "value": True,
                        },
                    ],
                }
            },
        )
        # Both conditions match
        ctx = {"user": {"plan": "enterprise"}, "account": {"region": "us"}}
        assert _evaluate_flag(flag, "staging", ctx) is True

        # Only one matches
        ctx2 = {"user": {"plan": "enterprise"}, "account": {"region": "eu"}}
        assert _evaluate_flag(flag, "staging", ctx2) is False

    def test_none_environment_returns_flag_default(self):
        flag = self._make_flag(default="default-val")
        assert _evaluate_flag(flag, None, {}) == "default-val"



"""Contains all the data models used in inputs/outputs"""

from .account import Account
from .account_resource import AccountResource
from .account_response import AccountResponse
from .api_key import ApiKey
from .api_key_data import ApiKeyData
from .api_key_list_response import ApiKeyListResponse
from .api_key_resource import ApiKeyResource
from .api_key_response import ApiKeyResponse
from .api_key_scopes import ApiKeyScopes
from .auth_token_response import AuthTokenResponse
from .context import Context
from .context_attributes import ContextAttributes
from .context_batch_response import ContextBatchResponse
from .context_bulk_item import ContextBulkItem
from .context_bulk_item_attributes import ContextBulkItemAttributes
from .context_bulk_register import ContextBulkRegister
from .context_list_response import ContextListResponse
from .context_resource import ContextResource
from .context_response import ContextResponse
from .context_type import ContextType
from .context_type_attributes import ContextTypeAttributes
from .context_type_list_response import ContextTypeListResponse
from .context_type_resource import ContextTypeResource
from .context_type_response import ContextTypeResponse
from .environment import Environment
from .environment_list_response import EnvironmentListResponse
from .environment_resource import EnvironmentResource
from .environment_response import EnvironmentResponse
from .error import Error
from .error_response import ErrorResponse
from .error_source_type_0 import ErrorSourceType0
from .invitation import Invitation
from .invitation_bulk_create_request import InvitationBulkCreateRequest
from .invitation_create_item import InvitationCreateItem
from .invitation_list_response import InvitationListResponse
from .invitation_resource import InvitationResource
from .invitation_response import InvitationResponse
from .limit_definition import LimitDefinition
from .login_request import LoginRequest
from .oidc_provider import OidcProvider
from .plan import Plan
from .plan_definition import PlanDefinition
from .plan_definition_limits import PlanDefinitionLimits
from .plan_list_response import PlanListResponse
from .plan_resource import PlanResource
from .product import Product
from .product_limits import ProductLimits
from .product_list_response import ProductListResponse
from .product_plans import ProductPlans
from .product_resource import ProductResource
from .register_request import RegisterRequest
from .service import Service
from .service_list_response import ServiceListResponse
from .service_resource import ServiceResource
from .service_response import ServiceResponse
from .user import User
from .user_list_response import UserListResponse
from .user_resource import UserResource
from .user_response import UserResponse

__all__ = (
    "Account",
    "AccountResource",
    "AccountResponse",
    "ApiKey",
    "ApiKeyData",
    "ApiKeyListResponse",
    "ApiKeyResource",
    "ApiKeyResponse",
    "ApiKeyScopes",
    "AuthTokenResponse",
    "Context",
    "ContextAttributes",
    "ContextBatchResponse",
    "ContextBulkItem",
    "ContextBulkItemAttributes",
    "ContextBulkRegister",
    "ContextListResponse",
    "ContextResource",
    "ContextResponse",
    "ContextType",
    "ContextTypeAttributes",
    "ContextTypeListResponse",
    "ContextTypeResource",
    "ContextTypeResponse",
    "Environment",
    "EnvironmentListResponse",
    "EnvironmentResource",
    "EnvironmentResponse",
    "Error",
    "ErrorResponse",
    "ErrorSourceType0",
    "Invitation",
    "InvitationBulkCreateRequest",
    "InvitationCreateItem",
    "InvitationListResponse",
    "InvitationResource",
    "InvitationResponse",
    "LimitDefinition",
    "LoginRequest",
    "OidcProvider",
    "Plan",
    "PlanDefinition",
    "PlanDefinitionLimits",
    "PlanListResponse",
    "PlanResource",
    "Product",
    "ProductLimits",
    "ProductListResponse",
    "ProductPlans",
    "ProductResource",
    "RegisterRequest",
    "Service",
    "ServiceListResponse",
    "ServiceResource",
    "ServiceResponse",
    "User",
    "UserListResponse",
    "UserResource",
    "UserResponse",
)



from collections.abc import Mapping
from typing import Any, TypeVar, TYPE_CHECKING

from attrs import define as _attrs_define
from attrs import field as _attrs_field


if TYPE_CHECKING:
    from ..models.service_resource import ServiceResource


T = TypeVar("T", bound="ServiceResponse")


@_attrs_define
class ServiceResponse:
    """
    Attributes:
        data (ServiceResource):  Example: {'attributes': {'created_at': '2026-03-20T11:02:16.616Z', 'key':
            'user_service', 'name': 'User Service', 'updated_at': '2026-03-20T11:02:16.616Z'}, 'id':
            'e5f6a7b8-c9d0-1234-efab-345678901234', 'type': 'service'}.
    """

    data: "ServiceResource"
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        data = self.data.to_dict()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "data": data,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.service_resource import ServiceResource

        d = dict(src_dict)
        data = ServiceResource.from_dict(d.pop("data"))

        service_response = cls(
            data=data,
        )

        service_response.additional_properties = d
        return service_response

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



"""Smpl Logging SDK module — wraps generated logging client."""

from __future__ import annotations

from smplkit.logging.client import (
    AsyncLoggingClient,
    LoggingClient,
    SmplLogGroup,
    SmplLogger,
)

__all__ = [
    "AsyncLoggingClient",
    "LoggingClient",
    "SmplLogGroup",
    "SmplLogger",
]



"""Top-level SDK clients — SmplClient (sync) and AsyncSmplClient (async)."""

from __future__ import annotations

import logging
import os

from smplkit._errors import SmplError
from smplkit._generated.app.api.contexts import bulk_register_contexts as gen_bulk_register_contexts
from smplkit._generated.app.models.context_bulk_item import ContextBulkItem
from smplkit._generated.app.models.context_bulk_item_attributes import ContextBulkItemAttributes
from smplkit._generated.app.models.context_bulk_register import ContextBulkRegister
from smplkit._generated.config.client import AuthenticatedClient
from smplkit._resolve import _resolve_api_key
from smplkit._ws import SharedWebSocket
from smplkit.config.client import AsyncConfigClient, ConfigClient
from smplkit.flags.client import AsyncFlagsClient, FlagsClient
from smplkit.logging.client import AsyncLoggingClient, LoggingClient

logger = logging.getLogger("smplkit")

_DEFAULT_BASE_URL = "https://config.smplkit.com"
_APP_BASE_URL = "https://app.smplkit.com"

_NO_API_KEY_MESSAGE = (
    "No API key provided. Set one of:\n"
    "  1. Pass api_key to the constructor\n"
    "  2. Set the SMPLKIT_API_KEY environment variable\n"
    "  3. Create a ~/.smplkit file with:\n"
    "     [default]\n"
    "     api_key = your_key_here"
)

_NO_ENVIRONMENT_MESSAGE = (
    "No environment provided. Set one of:\n"
    "  1. Pass environment to the constructor\n"
    "  2. Set the SMPLKIT_ENVIRONMENT environment variable"
)

_NOT_CONNECTED_MESSAGE = "SmplClient is not connected. Call client.connect() first."


class SmplClient:
    """Synchronous entry point for the smplkit SDK.

    Usage::

        from smplkit import SmplClient

        with SmplClient("sk_api_...", environment="production") as client:
            checkout_v2 = client.flags.boolFlag("checkout-v2", False)
            client.connect()
            if checkout_v2.get(): ...

    The API key is optional. When omitted, it is resolved from the
    ``SMPLKIT_API_KEY`` environment variable or the ``~/.smplkit``
    configuration file.

    Args:
        api_key: API key for authenticating with the smplkit platform.
            When *None*, the SDK resolves it automatically.
        environment: The environment to connect to (e.g. ``"production"``).
            Required — resolved from ``SMPLKIT_ENVIRONMENT`` if not provided.
        service: Optional service name. When set, the SDK automatically
            registers the service as a context instance and includes it
            in flag evaluation context.
    """

    def __init__(
        self,
        api_key: str | None = None,
        *,
        environment: str | None = None,
        service: str | None = None,
    ) -> None:
        resolved = _resolve_api_key(api_key)
        if resolved is None:
            raise SmplError(_NO_API_KEY_MESSAGE)
        self._api_key = resolved

        resolved_env = environment or os.environ.get("SMPLKIT_ENVIRONMENT")
        if not resolved_env:
            raise SmplError(_NO_ENVIRONMENT_MESSAGE)
        self._environment = resolved_env
        self._service = service or os.environ.get("SMPLKIT_SERVICE") or None

        self._http_client = AuthenticatedClient(
            base_url=_DEFAULT_BASE_URL,
            token=resolved,
        )
        self._app_http = AuthenticatedClient(
            base_url=_APP_BASE_URL,
            token=resolved,
        )
        self._ws_manager: SharedWebSocket | None = None
        self._connected = False
        self.config = ConfigClient(self)
        self.flags = FlagsClient(self)
        self.logging = LoggingClient(self)

    def connect(self) -> None:
        """Connect to the smplkit platform.

        Opens the shared WebSocket, fetches initial flag and config data,
        and registers the service as a context instance (if provided).

        This method is idempotent — calling it multiple times is safe.
        """
        if self._connected:
            return

        # Register service context (fire-and-forget)
        if self._service:
            self._register_service_context()

        # Connect flags (fetch definitions, register WS listeners)
        self.flags._connect_internal()

        # Connect config (fetch all, resolve, cache)
        self.config._connect_internal()

        # Connect logging (discover, register, fetch, resolve, apply)
        self.logging._connect_internal()

        self._connected = True

    def _register_service_context(self) -> None:
        """Register the service as a context instance on the app service."""
        try:
            attrs = ContextBulkItemAttributes()
            attrs.additional_properties = {"name": self._service}
            item = ContextBulkItem(type_="service", key=self._service, attributes=attrs)
            body = ContextBulkRegister(contexts=[item])
            gen_bulk_register_contexts.sync_detailed(client=self._app_http, body=body)
        except Exception:
            logger.warning("Failed to register service context", exc_info=True)

    def _ensure_ws(self) -> SharedWebSocket:
        """Lazily create and start the shared WebSocket."""
        if self._ws_manager is None:
            self._ws_manager = SharedWebSocket(
                app_base_url=_APP_BASE_URL,
                api_key=self._api_key,
            )
            self._ws_manager.start()
        return self._ws_manager

    def close(self) -> None:
        """Close the underlying HTTP connection pool and shared WebSocket."""
        self.logging._close()
        if self._ws_manager is not None:
            self._ws_manager.stop()
            self._ws_manager = None
        client = self._http_client._client
        if client is not None:
            client.close()
            self._http_client._client = None

    def __enter__(self) -> SmplClient:
        return self

    def __exit__(self, *args: object) -> None:
        self.close()


class AsyncSmplClient:
    """Asynchronous entry point for the smplkit SDK.

    Usage::

        from smplkit import AsyncSmplClient

        async with AsyncSmplClient("sk_api_...", environment="production") as client:
            checkout_v2 = client.flags.boolFlag("checkout-v2", False)
            await client.connect()
            if checkout_v2.get(): ...

    The API key is optional. When omitted, it is resolved from the
    ``SMPLKIT_API_KEY`` environment variable or the ``~/.smplkit``
    configuration file.

    Args:
        api_key: API key for authenticating with the smplkit platform.
            When *None*, the SDK resolves it automatically.
        environment: The environment to connect to (e.g. ``"production"``).
            Required — resolved from ``SMPLKIT_ENVIRONMENT`` if not provided.
        service: Optional service name. When set, the SDK automatically
            registers the service as a context instance and includes it
            in flag evaluation context.
    """

    def __init__(
        self,
        api_key: str | None = None,
        *,
        environment: str | None = None,
        service: str | None = None,
    ) -> None:
        resolved = _resolve_api_key(api_key)
        if resolved is None:
            raise SmplError(_NO_API_KEY_MESSAGE)
        self._api_key = resolved

        resolved_env = environment or os.environ.get("SMPLKIT_ENVIRONMENT")
        if not resolved_env:
            raise SmplError(_NO_ENVIRONMENT_MESSAGE)
        self._environment = resolved_env
        self._service = service or os.environ.get("SMPLKIT_SERVICE") or None

        self._http_client = AuthenticatedClient(
            base_url=_DEFAULT_BASE_URL,
            token=resolved,
        )
        self._app_http = AuthenticatedClient(
            base_url=_APP_BASE_URL,
            token=resolved,
        )
        self._ws_manager: SharedWebSocket | None = None
        self._connected = False
        self.config = AsyncConfigClient(self)
        self.flags = AsyncFlagsClient(self)
        self.logging = AsyncLoggingClient(self)

    async def connect(self) -> None:
        """Connect to the smplkit platform.

        Opens the shared WebSocket, fetches initial flag and config data,
        and registers the service as a context instance (if provided).

        This method is idempotent — calling it multiple times is safe.
        """
        if self._connected:
            return

        # Register service context (fire-and-forget)
        if self._service:
            await self._register_service_context()

        # Connect flags (fetch definitions, register WS listeners)
        await self.flags._connect_internal()

        # Connect config (fetch all, resolve, cache)
        await self.config._connect_internal()

        # Connect logging (discover, register, fetch, resolve, apply)
        await self.logging._connect_internal()

        self._connected = True

    async def _register_service_context(self) -> None:
        """Register the service as a context instance on the app service."""
        try:
            attrs = ContextBulkItemAttributes()
            attrs.additional_properties = {"name": self._service}
            item = ContextBulkItem(type_="service", key=self._service, attributes=attrs)
            body = ContextBulkRegister(contexts=[item])
            await gen_bulk_register_contexts.asyncio_detailed(client=self._app_http, body=body)
        except Exception:
            logger.warning("Failed to register service context", exc_info=True)

    def _ensure_ws(self) -> SharedWebSocket:
        """Lazily create and start the shared WebSocket."""
        if self._ws_manager is None:
            self._ws_manager = SharedWebSocket(
                app_base_url=_APP_BASE_URL,
                api_key=self._api_key,
            )
            self._ws_manager.start()
        return self._ws_manager

    async def close(self) -> None:
        """Close the underlying HTTP connection pool and shared WebSocket."""
        self.logging._close()
        if self._ws_manager is not None:
            self._ws_manager.stop()
            self._ws_manager = None
        client = self._http_client._async_client
        if client is not None:
            await client.aclose()
            self._http_client._async_client = None

    async def __aenter__(self) -> AsyncSmplClient:
        return self

    async def __aexit__(self, *args: object) -> None:
        await self.close()



"""
Smpl Config SDK Showcase
========================

Demonstrates the smplkit Python SDK for Smpl Config, covering:

- Client initialization (``AsyncSmplClient``)
- Management-plane CRUD: create, update, list, and delete configs
- Environment-specific overrides and multi-level inheritance
- Prescriptive value resolution via ``client.connect()`` + ``client.config.get()``
- Typed accessors: ``get_str()``, ``get_int()``, ``get_bool()``
- Manual refresh: ``client.config.refresh()``
- Change listeners: ``client.config.on_change()``

This script is designed to be read top-to-bottom as a walkthrough of the
SDK's full capability surface. It is runnable against a live smplkit
environment, but is *not* a test — it creates, modifies, and deletes
real configs.

Prerequisites:
    - ``pip install smplkit-sdk``
    - A valid smplkit API key (set via ``SMPLKIT_API_KEY`` env var)
    - The smplkit Config service running and reachable

Usage::

    export SMPLKIT_API_KEY="sk_api_..."
    export SMPLKIT_ENVIRONMENT="production"
    python examples/config_showcase.py
"""

import asyncio
import os
import sys

from smplkit import AsyncSmplClient

# ---------------------------------------------------------------------------
# Configuration
# ---------------------------------------------------------------------------

API_KEY = os.environ.get("SMPLKIT_API_KEY", "")
ENVIRONMENT = os.environ.get("SMPLKIT_ENVIRONMENT", "production")

if not API_KEY:
    print("ERROR: Set the SMPLKIT_API_KEY environment variable before running.")
    print("  export SMPLKIT_API_KEY='sk_api_...'")
    sys.exit(1)

# ---------------------------------------------------------------------------
# Helpers
# ---------------------------------------------------------------------------


def section(title: str) -> None:
    """Print a section header for readability."""
    print(f"\n{'=' * 60}")
    print(f"  {title}")
    print(f"{'=' * 60}\n")


def step(description: str) -> None:
    """Print a step within a section."""
    print(f"  → {description}")


async def main() -> None:

    # ======================================================================
    # 1. SDK INITIALIZATION
    # ======================================================================
    section("1. SDK Initialization")

    # The AsyncSmplClient is the entry point for async applications.
    # Environment is required — it determines how config values are resolved.
    client = AsyncSmplClient(API_KEY, environment=ENVIRONMENT)
    step(f"AsyncSmplClient initialized (environment={ENVIRONMENT})")

    # ======================================================================
    # 2. MANAGEMENT PLANE — Set up the configuration hierarchy
    # ======================================================================

    # ------------------------------------------------------------------
    # 2a. Update the built-in common config
    # ------------------------------------------------------------------
    section("2a. Update the Common Config")

    common = await client.config.get(key="common")
    step(f"Fetched common config: id={common.id}, key={common.key}")

    await common.update(
        description="Organization-wide shared configuration",
        items={
            "app_name": {"value": "Acme SaaS Platform", "type": "STRING"},
            "support_email": {"value": "support@acme.dev", "type": "STRING"},
            "max_retries": {"value": 3, "type": "NUMBER"},
            "request_timeout_ms": {"value": 5000, "type": "NUMBER"},
            "pagination_default_page_size": {"value": 25, "type": "NUMBER"},
        },
    )
    step("Common config base values set")

    await common.set_values(
        {"max_retries": 5, "request_timeout_ms": 10000},
        environment="production",
    )
    step("Common config production overrides set")

    await common.set_values(
        {"max_retries": 2},
        environment="staging",
    )
    step("Common config staging overrides set")

    # ------------------------------------------------------------------
    # 2b. Create a service-specific config (inherits from common)
    # ------------------------------------------------------------------
    section("2b. Create the User Service Config")

    user_service = await client.config.create(
        name="User Service",
        key="user_service",
        description="Configuration for the user microservice.",
        items={
            "database": {"value": {"host": "localhost", "port": 5432, "name": "users_dev", "pool_size": 5}},
            "cache_ttl_seconds": {"value": 300, "type": "NUMBER"},
            "enable_signup": {"value": True, "type": "BOOLEAN"},
            "pagination_default_page_size": {"value": 50, "type": "NUMBER"},
        },
    )
    step(f"Created user_service config: id={user_service.id}")

    await user_service.set_values(
        {
            "database": {"host": "prod-users-rds.internal.acme.dev", "name": "users_prod", "pool_size": 20},
            "cache_ttl_seconds": 600,
        },
        environment="production",
    )
    step("User service production overrides set")

    await user_service.set_value("enable_signup", False, environment="production")
    step("Disabled signup in production")

    # ------------------------------------------------------------------
    # 2c. Create a child config for multi-level inheritance
    # ------------------------------------------------------------------
    section("2c. Create the Auth Module Config (child of User Service)")

    auth_module = await client.config.create(
        name="Auth Module",
        key="auth_module",
        description="Authentication module within the user service.",
        parent=user_service.id,
        items={
            "session_ttl_minutes": {"value": 60, "type": "NUMBER"},
            "mfa_enabled": {"value": False, "type": "BOOLEAN"},
        },
    )
    step(f"Created auth_module config: id={auth_module.id}")

    await auth_module.set_values(
        {"session_ttl_minutes": 30, "mfa_enabled": True},
        environment="production",
    )
    step("Auth module production overrides set")

    # ------------------------------------------------------------------
    # 2d. List all configs
    # ------------------------------------------------------------------
    section("2d. List All Configs")

    configs = await client.config.list()
    for cfg in configs:
        parent_info = f" (parent: {cfg.parent})" if cfg.parent else " (root)"
        step(f"{cfg.key}{parent_info}")

    # ======================================================================
    # 3. PRESCRIPTIVE PLANE — Connect and read resolved values
    # ======================================================================
    #
    # client.connect() fetches all configs, resolves values for the
    # environment, and caches everything in-process. After connect(),
    # client.config.get() is a pure local dict read — zero network.
    #
    # ======================================================================

    section("3a. Connect and Read Resolved Values")

    await client.connect()
    step("client.connect() completed — all configs fetched and cached")

    # Prescriptive access: get(config_key, item_key)
    db_config = await client.config.get("user_service", "database")
    step(f"database = {db_config}")

    retries = await client.config.get("user_service", "max_retries")
    step(f"max_retries = {retries}")
    # Expected: 5 (inherited from common's production override)

    cache_ttl = await client.config.get("user_service", "cache_ttl_seconds")
    step(f"cache_ttl_seconds = {cache_ttl}")
    # Expected: 600 (user_service production override)

    page_size = await client.config.get("user_service", "pagination_default_page_size")
    step(f"pagination_default_page_size = {page_size}")
    # Expected: 50 (user_service base overrides common's 25)

    missing = await client.config.get("user_service", "nonexistent_key")
    step(f"nonexistent key = {missing}")
    # Expected: None

    # Get all values for a config as a dict
    all_values = await client.config.get("user_service")
    step(f"Total resolved keys for user_service: {len(all_values)}")

    # ------------------------------------------------------------------
    # 3b. Typed accessors
    # ------------------------------------------------------------------
    section("3b. Typed Accessors")

    app_name = await client.config.get_str("user_service", "app_name", default="Unknown")
    step(f"app_name (str) = {app_name}")

    timeout_ms = await client.config.get_int("user_service", "request_timeout_ms", default=3000)
    step(f"request_timeout_ms (int) = {timeout_ms}")

    signup = await client.config.get_bool("user_service", "enable_signup", default=True)
    step(f"enable_signup (bool) = {signup}")
    # Expected: False (production override)

    # ------------------------------------------------------------------
    # 3c. Multi-level inheritance
    # ------------------------------------------------------------------
    section("3c. Multi-Level Inheritance (auth_module)")

    session_ttl = await client.config.get("auth_module", "session_ttl_minutes")
    step(f"session_ttl_minutes = {session_ttl}")
    # Expected: 30 (auth_module production override)

    mfa = await client.config.get("auth_module", "mfa_enabled")
    step(f"mfa_enabled = {mfa}")
    # Expected: True (auth_module production override)

    inherited_app = await client.config.get("auth_module", "app_name")
    step(f"app_name (inherited from common) = {inherited_app}")

    # ======================================================================
    # 4. CHANGE LISTENERS AND REFRESH
    # ======================================================================

    section("4a. Change Listeners")

    changes: list = []

    def on_change(event) -> None:
        changes.append(event)
        print(f"    [CHANGE] {event.config_key}.{event.item_key}: "
              f"{event.old_value!r} -> {event.new_value!r}")

    client.config.on_change(on_change)
    step("Global change listener registered")

    retries_changes: list = []
    client.config.on_change(
        lambda e: retries_changes.append(e),
        config_key="common",
        item_key="max_retries",
    )
    step("Key-specific listener registered for common.max_retries")

    # ------------------------------------------------------------------
    # 4b. Trigger a change via management API, then refresh
    # ------------------------------------------------------------------
    section("4b. Refresh After Management Change")

    await common.set_value("max_retries", 7, environment="production")
    step("Updated max_retries to 7 on common (production)")

    await client.config.refresh()
    step("client.config.refresh() completed")

    new_retries = await client.config.get("user_service", "max_retries")
    step(f"max_retries after refresh = {new_retries}")
    # Expected: 7

    step(f"Global changes received: {len(changes)}")
    step(f"Retries-specific changes received: {len(retries_changes)}")

    # ======================================================================
    # 5. SYNC CLIENT DEMO
    # ======================================================================
    section("5. Sync Client (same API, no await)")

    # For sync applications (Django, Flask, CLI tools):
    #
    #     from smplkit import SmplClient
    #
    #     client = SmplClient("sk_api_...", environment="production")
    #     client.connect()
    #     host = client.config.get_str("user_service", "database_host")
    #     retries = client.config.get_int("user_service", "max_retries", default=3)
    #     client.config.refresh()  # re-fetch all configs
    #     client.close()

    step("(See code comments for sync usage examples)")

    # ======================================================================
    # 6. CLEANUP
    # ======================================================================
    section("6. Cleanup")

    await client.config.delete(auth_module.id)
    step(f"Deleted auth_module ({auth_module.id})")

    await client.config.delete(user_service.id)
    step(f"Deleted user_service ({user_service.id})")

    await common.update(description="", items={}, environments={})
    step("Common config reset to empty")

    await client.close()
    step("AsyncSmplClient closed")

    # ======================================================================
    # DONE
    # ======================================================================
    section("ALL DONE")
    print("  The Config SDK showcase completed successfully.\n")


if __name__ == "__main__":
    asyncio.run(main())



"""Tests for logger name normalization."""

from smplkit.logging._normalize import normalize_logger_name


class TestNormalizeLoggerName:
    def test_slash_replaced_with_dot(self):
        assert normalize_logger_name("a/b/c") == "a.b.c"

    def test_colon_replaced_with_dot(self):
        assert normalize_logger_name("a:b:c") == "a.b.c"

    def test_lowercased(self):
        assert normalize_logger_name("ABC") == "abc"

    def test_combined(self):
        assert normalize_logger_name("ABC/Def:Ghi") == "abc.def.ghi"

    def test_already_normalized(self):
        assert normalize_logger_name("com.example.sql") == "com.example.sql"

    def test_mixed_separators(self):
        assert normalize_logger_name("APP/Module:Sub.Name") == "app.module.sub.name"

    def test_empty_string(self):
        assert normalize_logger_name("") == ""

    def test_single_name(self):
        assert normalize_logger_name("ROOT") == "root"



"""Tests for helper functions, FlagChangeEvent, and remaining uncovered code."""

from unittest.mock import MagicMock, patch

import pytest

from smplkit._errors import SmplConflictError, SmplNotFoundError, SmplValidationError
from smplkit.flags.client import (
    FlagChangeEvent,
    FlagsClient,
    _ContextRegistrationBuffer,
    _CONTEXT_REGISTRATION_LRU_SIZE,
    _FlagHandle,
    _build_gen_flag,
    _check_response_status,
    _evaluate_flag,
    _extract_environments,
    _extract_rule,
    _extract_values,
    _maybe_reraise_network_error,
    _unset_to_none,
)
from smplkit.flags.types import Context


# ---------------------------------------------------------------------------
# _check_response_status
# ---------------------------------------------------------------------------


class TestCheckResponseStatus:
    def test_409_raises_conflict(self):
        with pytest.raises(SmplConflictError):
            _check_response_status(409, b"conflict detail")

    def test_422_raises_validation(self):
        with pytest.raises(SmplValidationError):
            _check_response_status(422, b"validation detail")

    def test_200_does_nothing(self):
        _check_response_status(200, b"ok")


# ---------------------------------------------------------------------------
# _maybe_reraise_network_error
# ---------------------------------------------------------------------------


class TestMaybeReraiseNetworkError:
    def test_reraises_not_found(self):
        with pytest.raises(SmplNotFoundError):
            _maybe_reraise_network_error(SmplNotFoundError("nope"))

    def test_reraises_validation(self):
        with pytest.raises(SmplValidationError):
            _maybe_reraise_network_error(SmplValidationError("bad"))

    def test_ignores_generic_exception(self):
        _maybe_reraise_network_error(ValueError("other"))


# ---------------------------------------------------------------------------
# _extract_environments / _extract_rule / _extract_values / _unset_to_none
# ---------------------------------------------------------------------------


class TestExtractEnvironments:
    def test_none_returns_empty(self):
        assert _extract_environments(None) == {}

    def test_dict_passthrough(self):
        d = {"prod": {"enabled": True}}
        assert _extract_environments(d) == d

    def test_unset_returns_empty(self):
        from smplkit._generated.flags.types import UNSET

        assert _extract_environments(UNSET) == {}

    def test_unset_by_type_name(self):
        """Cover line 89: type_name == 'Unset' fallback check."""

        class Unset:
            pass

        assert _extract_environments(Unset()) == {}

    def test_unknown_type_returns_empty(self):
        """Cover line 108: fallthrough return for unknown type."""
        assert _extract_environments(42) == {}

    def test_gen_flag_environments(self):
        from smplkit._generated.flags.models.flag_environment import FlagEnvironment
        from smplkit._generated.flags.models.flag_environments import FlagEnvironments
        from smplkit._generated.flags.models.flag_rule import FlagRule
        from smplkit._generated.flags.models.flag_rule_logic import FlagRuleLogic

        logic = FlagRuleLogic()
        logic.additional_properties = {"==": [{"var": "user.plan"}, "enterprise"]}
        rule = FlagRule(logic=logic, value=True, description="test rule")
        env = FlagEnvironment(enabled=True, default=False, rules=[rule])
        envs = FlagEnvironments()
        envs.additional_properties = {"production": env}

        result = _extract_environments(envs)
        assert "production" in result
        assert result["production"]["enabled"] is True
        assert result["production"]["default"] is False
        assert len(result["production"]["rules"]) == 1
        assert result["production"]["rules"][0]["description"] == "test rule"

    def test_gen_flag_environments_unset_fields(self):
        from smplkit._generated.flags.models.flag_environment import FlagEnvironment
        from smplkit._generated.flags.models.flag_environments import FlagEnvironments
        from smplkit._generated.flags.types import UNSET

        env = FlagEnvironment(enabled=UNSET, default=UNSET, rules=UNSET)
        envs = FlagEnvironments()
        envs.additional_properties = {"staging": env}

        result = _extract_environments(envs)
        assert "staging" in result
        assert result["staging"]["rules"] == []


class TestExtractRule:
    def test_basic_rule(self):
        from smplkit._generated.flags.models.flag_rule import FlagRule
        from smplkit._generated.flags.models.flag_rule_logic import FlagRuleLogic

        logic = FlagRuleLogic()
        logic.additional_properties = {"==": [1, 1]}
        rule = FlagRule(logic=logic, value="enabled", description="always on")
        result = _extract_rule(rule)
        assert result["logic"] == {"==": [1, 1]}
        assert result["value"] == "enabled"
        assert result["description"] == "always on"

    def test_rule_without_description(self):
        from smplkit._generated.flags.models.flag_rule import FlagRule
        from smplkit._generated.flags.models.flag_rule_logic import FlagRuleLogic
        from smplkit._generated.flags.types import UNSET

        logic = FlagRuleLogic()
        logic.additional_properties = {}
        rule = FlagRule(logic=logic, value=42, description=UNSET)
        result = _extract_rule(rule)
        assert "description" not in result


class TestExtractValues:
    def test_empty_returns_empty(self):
        assert _extract_values([]) == []
        assert _extract_values(None) == []

    def test_extracts_values(self):
        v1 = MagicMock()
        v1.name = "On"
        v1.value = True
        result = _extract_values([v1])
        assert result == [{"name": "On", "value": True}]


class TestUnsetToNone:
    def test_none_passthrough(self):
        assert _unset_to_none(None) is None

    def test_string_passthrough(self):
        assert _unset_to_none("hello") == "hello"

    def test_unset_becomes_none(self):
        from smplkit._generated.flags.types import UNSET

        assert _unset_to_none(UNSET) is None


# ---------------------------------------------------------------------------
# _build_gen_flag
# ---------------------------------------------------------------------------


class TestBuildGenFlag:
    def test_with_environments(self):
        result = _build_gen_flag(
            key="test",
            name="Test",
            type_="BOOLEAN",
            default=False,
            values=[{"name": "True", "value": True}],
            environments={
                "production": {
                    "enabled": True,
                    "default": False,
                    "rules": [{"logic": {"==": [1, 1]}, "value": True, "description": "always"}],
                }
            },
        )
        assert result.key == "test"
        assert "production" in result.environments.additional_properties

    def test_without_environments(self):
        result = _build_gen_flag(
            key="test",
            name="Test",
            type_="STRING",
            default="off",
            values=[],
        )
        assert result.key == "test"


# ---------------------------------------------------------------------------
# FlagChangeEvent
# ---------------------------------------------------------------------------


class TestFlagChangeEvent:
    def test_construction(self):
        event = FlagChangeEvent(key="my-flag", source="websocket")
        assert event.key == "my-flag"
        assert event.source == "websocket"

    def test_repr(self):
        event = FlagChangeEvent(key="f", source="manual")
        assert "FlagChangeEvent" in repr(event)
        assert "f" in repr(event)


# ---------------------------------------------------------------------------
# _FlagHandle base class
# ---------------------------------------------------------------------------


class TestFlagHandleBase:
    def test_base_get(self):
        ns = MagicMock()
        ns._evaluate_handle.return_value = 42
        handle = _FlagHandle(ns, "test", 0)
        assert handle.get() == 42


# ---------------------------------------------------------------------------
# _ContextRegistrationBuffer LRU eviction
# ---------------------------------------------------------------------------


class TestContextBufferLRU:
    def test_lru_eviction(self):
        buf = _ContextRegistrationBuffer()
        # Fill beyond LRU size
        for i in range(_CONTEXT_REGISTRATION_LRU_SIZE + 1):
            buf.observe([Context("user", f"u-{i}")])
        # The seen dict should be capped
        assert len(buf._seen) == _CONTEXT_REGISTRATION_LRU_SIZE


# ---------------------------------------------------------------------------
# FlagsClient handle factories (via mock parent)
# ---------------------------------------------------------------------------


class TestFlagsClientHandleFactories:
    def _make_client(self):
        parent = MagicMock()
        parent._api_key = "test-key"
        with patch("smplkit.flags.client.AuthenticatedClient"):
            return FlagsClient(parent)

    def test_string_flag(self):
        client = self._make_client()
        handle = client.stringFlag("color", "red")
        assert handle.key == "color"
        assert handle.default == "red"

    def test_number_flag(self):
        client = self._make_client()
        handle = client.numberFlag("retries", 3)
        assert handle.key == "retries"
        assert handle.default == 3

    def test_json_flag(self):
        client = self._make_client()
        handle = client.jsonFlag("config", {"a": 1})
        assert handle.key == "config"

    def test_connection_status_disconnected(self):
        client = self._make_client()
        assert client.connection_status() == "disconnected"


# ---------------------------------------------------------------------------
# _evaluate_flag edge cases
# ---------------------------------------------------------------------------


class TestEvaluateFlagEdgeCases:
    def test_json_logic_error_returns_fallback(self):
        flag_def = {
            "key": "test",
            "default": "fallback",
            "environments": {
                "prod": {
                    "enabled": True,
                    "default": "env-default",
                    "rules": [{"logic": {"invalid_op": [1, 2]}, "value": "matched"}],
                }
            },
        }
        result = _evaluate_flag(flag_def, "prod", {})
        # Invalid logic should be caught, fall through to env default
        assert result == "env-default"

    def test_empty_logic_skipped(self):
        """Cover line 1420: empty logic dict causes continue."""
        flag_def = {
            "key": "test",
            "default": "fallback",
            "environments": {
                "prod": {
                    "enabled": True,
                    "default": "env-default",
                    "rules": [{"logic": {}, "value": "never"}],
                }
            },
        }
        result = _evaluate_flag(flag_def, "prod", {})
        assert result == "env-default"


# ---------------------------------------------------------------------------
# Context.__repr__
# ---------------------------------------------------------------------------


class TestContextRepr:
    def test_repr(self):
        ctx = Context("user", "u-1", name="Alice", plan="enterprise")
        r = repr(ctx)
        assert "user" in r
        assert "u-1" in r
        assert "Alice" in r



"""Tests for Flag and AsyncFlag resource models."""

import asyncio
from unittest.mock import MagicMock, AsyncMock

import pytest

from smplkit.flags.models import AsyncFlag, ContextType, Flag


class TestFlag:
    def _make_flag(self, **overrides):
        defaults = {
            "id": "flag-id-1",
            "key": "test-flag",
            "name": "Test Flag",
            "type": "BOOLEAN",
            "default": False,
            "values": [{"name": "True", "value": True}],
            "description": "A test flag",
            "environments": {"staging": {"enabled": True, "rules": []}},
        }
        defaults.update(overrides)
        client = MagicMock()
        return Flag(client, **defaults)

    def test_properties(self):
        flag = self._make_flag()
        assert flag.id == "flag-id-1"
        assert flag.key == "test-flag"
        assert flag.name == "Test Flag"
        assert flag.type == "BOOLEAN"
        assert flag.default is False
        assert flag.description == "A test flag"

    def test_repr(self):
        flag = self._make_flag()
        assert "test-flag" in repr(flag)

    def test_update_calls_client(self):
        client = MagicMock()
        flag = Flag(
            client,
            id="id-1",
            key="k",
            name="n",
            type="BOOLEAN",
            default=False,
            values=[],
        )
        updated = Flag(
            client,
            id="id-1",
            key="k",
            name="Updated",
            type="BOOLEAN",
            default=True,
            values=[],
        )
        client._update_flag.return_value = updated
        flag.update(name="Updated", default=True)
        assert flag.name == "Updated"
        assert flag.default is True

    def test_addRule_requires_environment(self):
        flag = self._make_flag()
        with pytest.raises(ValueError, match="environment"):
            flag.addRule({"logic": {}, "value": True})

    def test_addRule_calls_client(self):
        client = MagicMock()
        flag = Flag(
            client,
            id="id-1",
            key="k",
            name="n",
            type="BOOLEAN",
            default=False,
            values=[],
            environments={"staging": {"enabled": True, "rules": []}},
        )
        # get returns a copy of the flag
        client.get.return_value = flag
        updated = Flag(
            client,
            id="id-1",
            key="k",
            name="n",
            type="BOOLEAN",
            default=False,
            values=[],
            environments={"staging": {"enabled": True, "rules": [{"logic": {}, "value": True}]}},
        )
        client._update_flag.return_value = updated
        flag.addRule({"environment": "staging", "logic": {}, "value": True})
        assert client._update_flag.called

    def test_environments_default_empty(self):
        client = MagicMock()
        flag = Flag(client, id="id-1", key="k", name="n", type="BOOLEAN", default=False, values=[])
        assert flag.environments == {}


class TestAsyncFlag:
    def _make_flag(self, **overrides):
        defaults = {
            "id": "flag-id-1",
            "key": "test-flag",
            "name": "Test Flag",
            "type": "BOOLEAN",
            "default": False,
            "values": [{"name": "True", "value": True}],
        }
        defaults.update(overrides)
        client = MagicMock()
        return AsyncFlag(client, **defaults)

    def test_repr(self):
        flag = self._make_flag()
        assert "test-flag" in repr(flag)

    def test_update_calls_client(self):
        client = AsyncMock()
        flag = AsyncFlag(
            client,
            id="id-1",
            key="k",
            name="n",
            type="BOOLEAN",
            default=False,
            values=[],
        )
        updated = AsyncFlag(
            client,
            id="id-1",
            key="k",
            name="Updated",
            type="BOOLEAN",
            default=True,
            values=[],
        )
        client._update_flag.return_value = updated

        async def _run():
            await flag.update(name="Updated", default=True)
            assert flag.name == "Updated"

        asyncio.run(_run())

    def test_addRule_requires_environment(self):
        flag = self._make_flag()

        async def _run():
            with pytest.raises(ValueError, match="environment"):
                await flag.addRule({"logic": {}, "value": True})

        asyncio.run(_run())

    def test_addRule_calls_client(self):
        client = AsyncMock()
        flag = AsyncFlag(
            client,
            id="id-1",
            key="k",
            name="n",
            type="BOOLEAN",
            default=False,
            values=[],
            environments={"staging": {"enabled": True, "rules": []}},
        )
        client.get.return_value = flag
        updated = AsyncFlag(
            client,
            id="id-1",
            key="k",
            name="n",
            type="BOOLEAN",
            default=False,
            values=[],
            environments={"staging": {"enabled": True, "rules": [{"logic": {}, "value": True}]}},
        )
        client._update_flag.return_value = updated

        async def _run():
            await flag.addRule({"environment": "staging", "logic": {}, "value": True})
            assert client._update_flag.called

        asyncio.run(_run())


class TestContextType:
    def test_properties(self):
        ct = ContextType(id="ct-1", key="user", name="User", attributes={"plan": {}})
        assert ct.id == "ct-1"
        assert ct.key == "user"
        assert ct.name == "User"
        assert ct.attributes == {"plan": {}}

    def test_default_attributes(self):
        ct = ContextType(id="ct-1", key="user", name="User")
        assert ct.attributes == {}

    def test_repr(self):
        ct = ContextType(id="ct-1", key="user", name="User")
        assert "user" in repr(ct)



"""Level resolution algorithm per ADR-034 §3.1."""

from __future__ import annotations

from typing import Any

_FALLBACK_LEVEL = "INFO"


def resolve_level(
    logger_key: str,
    environment: str,
    loggers: dict[str, dict[str, Any]],
    groups: dict[str, dict[str, Any]],
) -> str:
    """Resolve the effective log level for a logger in an environment.

    Resolution chain (first non-null wins):

    1. Logger's own ``environments[env].level``
    2. Logger's own ``level``
    3. Group chain (recursive: group's env level -> group's level -> parent group...)
    4. Dot-notation ancestry (walk ``com.acme.payments`` -> ``com.acme`` -> ``com``,
       applying steps 1-3 at each)
    5. System fallback: ``"INFO"``

    Args:
        logger_key: The normalized logger key.
        environment: The current environment name.
        loggers: Dict of all loggers keyed by ``key``.
        groups: Dict of all log groups keyed by ``id``.

    Returns:
        The resolved smplkit level string.
    """
    result = _resolve_for_entry(logger_key, environment, loggers, groups)
    if result is not None:
        return result

    # Dot-notation ancestry: walk up the hierarchy
    parts = logger_key.split(".")
    for i in range(len(parts) - 1, 0, -1):
        ancestor_key = ".".join(parts[:i])
        result = _resolve_for_entry(ancestor_key, environment, loggers, groups)
        if result is not None:
            return result

    return _FALLBACK_LEVEL


def _resolve_for_entry(
    key: str,
    environment: str,
    loggers: dict[str, dict[str, Any]],
    groups: dict[str, dict[str, Any]],
) -> str | None:
    """Try to resolve a level for a single entry (logger or ancestor)."""
    entry = loggers.get(key)
    if entry is None:
        return None

    # Step 1: env override on the entry itself
    env_level = _env_level(entry, environment)
    if env_level is not None:
        return env_level

    # Step 2: base level on the entry itself
    base = entry.get("level")
    if base is not None:
        return base

    # Step 3: group chain
    group_id = entry.get("group")
    return _resolve_group_chain(group_id, environment, groups)


def _resolve_group_chain(
    group_id: str | None,
    environment: str,
    groups: dict[str, dict[str, Any]],
) -> str | None:
    """Walk the group chain looking for a level."""
    visited: set[str] = set()
    current_id = group_id
    while current_id is not None and current_id not in visited:
        visited.add(current_id)
        group = groups.get(current_id)
        if group is None:
            break
        env_level = _env_level(group, environment)
        if env_level is not None:
            return env_level
        base = group.get("level")
        if base is not None:
            return base
        current_id = group.get("group")
    return None


def _env_level(entry: dict[str, Any], environment: str) -> str | None:
    """Extract the environment-specific level from an entry, if present."""
    envs = entry.get("environments")
    if envs and isinstance(envs, dict):
        env_data = envs.get(environment)
        if env_data and isinstance(env_data, dict):
            level = env_data.get("level")
            if level is not None:
                return level
    return None



"""Tests for prescriptive config access: typed accessors, refresh, change listeners."""

import asyncio
from unittest.mock import MagicMock, patch

import pytest

from smplkit._errors import SmplNotConnectedError
from smplkit.client import AsyncSmplClient, SmplClient
from smplkit.config.client import ConfigChangeEvent


# ---------------------------------------------------------------------------
# Helpers
# ---------------------------------------------------------------------------


def _make_connected_client(cache: dict | None = None):
    """Create a SmplClient with a pre-populated config cache."""
    client = SmplClient(api_key="sk_test", environment="production")
    if cache is None:
        cache = {
            "db": {
                "host": "localhost",
                "port": 5432,
                "retries": 3,
                "enabled": True,
                "name": "test",
                "ratio": 0.75,
            },
        }
    client.config._config_cache = cache
    client.config._connected = True
    return client


def _make_connected_async_client(cache: dict | None = None):
    """Create an AsyncSmplClient with a pre-populated config cache."""
    client = AsyncSmplClient(api_key="sk_test", environment="production")
    if cache is None:
        cache = {
            "db": {
                "host": "localhost",
                "port": 5432,
                "retries": 3,
                "enabled": True,
                "name": "test",
                "ratio": 0.75,
            },
        }
    client.config._config_cache = cache
    client.config._connected = True
    return client


# ===================================================================
# 1. Typed accessors — sync
# ===================================================================


class TestTypedAccessorsSync:
    def test_get_str_returns_string(self):
        client = _make_connected_client()
        assert client.config.get_str("db", "name") == "test"

    def test_get_str_returns_default_for_non_string(self):
        client = _make_connected_client()
        assert client.config.get_str("db", "port", default="nope") == "nope"

    def test_get_str_returns_default_for_missing(self):
        client = _make_connected_client()
        assert client.config.get_str("db", "ghost") is None

    def test_get_int_returns_int(self):
        client = _make_connected_client()
        assert client.config.get_int("db", "port") == 5432

    def test_get_int_returns_default_for_non_int(self):
        client = _make_connected_client()
        assert client.config.get_int("db", "name", default=0) == 0

    def test_get_int_returns_default_for_bool(self):
        client = _make_connected_client()
        assert client.config.get_int("db", "enabled", default=-1) == -1

    def test_get_bool_returns_bool(self):
        client = _make_connected_client()
        assert client.config.get_bool("db", "enabled") is True

    def test_get_bool_returns_default_for_non_bool(self):
        client = _make_connected_client()
        assert client.config.get_bool("db", "port", default=False) is False

    def test_get_float_returns_float(self):
        client = _make_connected_client()
        assert client.config.get_float("db", "ratio") == 0.75

    def test_get_float_promotes_int(self):
        client = _make_connected_client()
        assert client.config.get_float("db", "port") == 5432.0

    def test_get_float_returns_default_for_bool(self):
        client = _make_connected_client()
        assert client.config.get_float("db", "enabled", default=0.0) == 0.0

    def test_get_float_returns_default_for_string(self):
        client = _make_connected_client()
        assert client.config.get_float("db", "name", default=1.0) == 1.0

    def test_typed_accessors_require_connect(self):
        client = SmplClient(api_key="sk_test", environment="test")
        with pytest.raises(SmplNotConnectedError):
            client.config.get_str("db", "host")
        with pytest.raises(SmplNotConnectedError):
            client.config.get_int("db", "port")
        with pytest.raises(SmplNotConnectedError):
            client.config.get_bool("db", "enabled")


# ===================================================================
# 2. Typed accessors — async
# ===================================================================


class TestTypedAccessorsAsync:
    def test_get_str_returns_string(self):
        client = _make_connected_async_client()
        assert asyncio.run(client.config.get_str("db", "name")) == "test"

    def test_get_str_returns_default_for_non_string(self):
        client = _make_connected_async_client()
        assert asyncio.run(client.config.get_str("db", "port", default="nope")) == "nope"

    def test_get_int_returns_int(self):
        client = _make_connected_async_client()
        assert asyncio.run(client.config.get_int("db", "port")) == 5432

    def test_get_int_returns_default_for_bool(self):
        client = _make_connected_async_client()
        assert asyncio.run(client.config.get_int("db", "enabled", default=-1)) == -1

    def test_get_bool_returns_bool(self):
        client = _make_connected_async_client()
        assert asyncio.run(client.config.get_bool("db", "enabled")) is True

    def test_get_bool_returns_default_for_non_bool(self):
        client = _make_connected_async_client()
        assert asyncio.run(client.config.get_bool("db", "port", default=False)) is False

    def test_get_float_returns_float(self):
        client = _make_connected_async_client()
        assert asyncio.run(client.config.get_float("db", "ratio")) == 0.75

    def test_get_float_promotes_int(self):
        client = _make_connected_async_client()
        assert asyncio.run(client.config.get_float("db", "port")) == 5432.0

    def test_get_float_returns_default_for_bool(self):
        client = _make_connected_async_client()
        assert asyncio.run(client.config.get_float("db", "enabled", default=0.0)) == 0.0

    def test_get_float_returns_default_for_string(self):
        client = _make_connected_async_client()
        assert asyncio.run(client.config.get_float("db", "name", default=1.0)) == 1.0


# ===================================================================
# 3. Refresh — sync
# ===================================================================


class TestRefreshSync:
    def test_refresh_requires_connect(self):
        client = SmplClient(api_key="sk_test", environment="test")
        with pytest.raises(SmplNotConnectedError):
            client.config.refresh()

    @patch("smplkit.config.client.list_configs.sync_detailed")
    def test_refresh_updates_cache(self, mock_list):
        client = _make_connected_client({"db": {"host": "old"}})

        # Mock the list response to return a config with new values
        mock_attrs = MagicMock()
        mock_attrs.name = "DB"
        mock_attrs.key = "db"
        mock_attrs.description = None
        mock_attrs.parent = None
        mock_attrs.items = None
        mock_attrs.environments = None
        mock_attrs.created_at = None
        mock_attrs.updated_at = None

        mock_resource = MagicMock()
        mock_resource.id = "cfg-1"
        mock_resource.attributes = mock_attrs

        mock_parsed = MagicMock()
        mock_parsed.data = [mock_resource]

        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.content = b""
        mock_response.parsed = mock_parsed
        mock_list.return_value = mock_response

        # The Config._build_chain is called during refresh
        with patch(
            "smplkit.config.models.Config._build_chain",
            return_value=[{"id": "cfg-1", "items": {}, "values": {"host": "new-host"}, "environments": {}}],
        ):
            client.config.refresh()

        assert client.config.get("db", "host") == "new-host"

    @patch("smplkit.config.client.list_configs.sync_detailed")
    def test_refresh_fires_listeners(self, mock_list):
        client = _make_connected_client({"db": {"host": "old"}})

        events = []
        client.config.on_change(lambda e: events.append(e))

        mock_attrs = MagicMock()
        mock_attrs.name = "DB"
        mock_attrs.key = "db"
        mock_attrs.description = None
        mock_attrs.parent = None
        mock_attrs.items = None
        mock_attrs.environments = None
        mock_attrs.created_at = None
        mock_attrs.updated_at = None

        mock_resource = MagicMock()
        mock_resource.id = "cfg-1"
        mock_resource.attributes = mock_attrs

        mock_parsed = MagicMock()
        mock_parsed.data = [mock_resource]

        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.content = b""
        mock_response.parsed = mock_parsed
        mock_list.return_value = mock_response

        with patch(
            "smplkit.config.models.Config._build_chain",
            return_value=[{"id": "cfg-1", "values": {"host": "new-host"}, "environments": {}}],
        ):
            client.config.refresh()

        assert len(events) == 1
        assert events[0].config_key == "db"
        assert events[0].item_key == "host"
        assert events[0].old_value == "old"
        assert events[0].new_value == "new-host"
        assert events[0].source == "manual"


# ===================================================================
# 4. Change listeners
# ===================================================================


class TestChangeListeners:
    def test_on_change_registers_listener(self):
        client = _make_connected_client()
        client.config.on_change(lambda e: None)
        assert len(client.config._listeners) == 1

    def test_on_change_with_config_key_filter(self):
        client = _make_connected_client()
        client.config.on_change(lambda e: None, config_key="db")
        assert client.config._listeners[0][1] == "db"

    def test_on_change_with_item_key_filter(self):
        client = _make_connected_client()
        client.config.on_change(lambda e: None, item_key="host")
        assert client.config._listeners[0][2] == "host"

    def test_fire_change_listeners_filters_by_config_key(self):
        client = _make_connected_client()
        events = []
        client.config.on_change(lambda e: events.append(e), config_key="db")
        client.config._fire_change_listeners(
            {"db": {"host": "old"}, "other": {"x": 1}},
            {"db": {"host": "new"}, "other": {"x": 2}},
            source="manual",
        )
        assert len(events) == 1
        assert events[0].config_key == "db"

    def test_fire_change_listeners_filters_by_item_key(self):
        client = _make_connected_client()
        events = []
        client.config.on_change(lambda e: events.append(e), item_key="host")
        client.config._fire_change_listeners(
            {"db": {"host": "old", "port": 1}},
            {"db": {"host": "new", "port": 2}},
            source="manual",
        )
        assert len(events) == 1
        assert events[0].item_key == "host"

    def test_no_change_fires_nothing(self):
        client = _make_connected_client()
        events = []
        client.config.on_change(lambda e: events.append(e))
        client.config._fire_change_listeners(
            {"db": {"host": "same"}},
            {"db": {"host": "same"}},
            source="manual",
        )
        assert len(events) == 0

    def test_listener_exception_is_caught(self):
        client = _make_connected_client()
        good_events = []

        def bad_listener(event):
            raise ValueError("boom")

        client.config.on_change(bad_listener)
        client.config.on_change(lambda e: good_events.append(e))

        client.config._fire_change_listeners(
            {"db": {"host": "old"}},
            {"db": {"host": "new"}},
            source="manual",
        )
        assert len(good_events) == 1


# ===================================================================
# 5. ConfigChangeEvent
# ===================================================================


class TestConfigChangeEvent:
    def test_attributes(self):
        event = ConfigChangeEvent(
            config_key="db",
            item_key="host",
            old_value="old",
            new_value="new",
            source="manual",
        )
        assert event.config_key == "db"
        assert event.item_key == "host"
        assert event.old_value == "old"
        assert event.new_value == "new"
        assert event.source == "manual"

    def test_repr(self):
        event = ConfigChangeEvent(
            config_key="db",
            item_key="host",
            old_value="old",
            new_value="new",
            source="websocket",
        )
        r = repr(event)
        assert "ConfigChangeEvent" in r
        assert "db" in r
        assert "host" in r


# ===================================================================
# 6. Async refresh and change listeners
# ===================================================================


class TestAsyncRefresh:
    def test_refresh_requires_connect(self):
        client = AsyncSmplClient(api_key="sk_test", environment="test")
        with pytest.raises(SmplNotConnectedError):
            asyncio.run(client.config.refresh())

    @patch("smplkit.config.client.list_configs.asyncio_detailed")
    def test_refresh_updates_cache(self, mock_list):
        client = _make_connected_async_client({"db": {"host": "old"}})

        mock_attrs = MagicMock()
        mock_attrs.name = "DB"
        mock_attrs.key = "db"
        mock_attrs.description = None
        mock_attrs.parent = None
        mock_attrs.items = None
        mock_attrs.environments = None
        mock_attrs.created_at = None
        mock_attrs.updated_at = None

        mock_resource = MagicMock()
        mock_resource.id = "cfg-1"
        mock_resource.attributes = mock_attrs

        mock_parsed = MagicMock()
        mock_parsed.data = [mock_resource]

        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.content = b""
        mock_response.parsed = mock_parsed

        async def fake_list(*args, **kwargs):
            return mock_response

        mock_list.side_effect = fake_list

        async def run():
            with patch(
                "smplkit.config.models.AsyncConfig._build_chain",
                return_value=[{"id": "cfg-1", "values": {"host": "new-host"}, "environments": {}}],
            ):
                await client.config.refresh()

        asyncio.run(run())
        assert client.config._config_cache["db"]["host"] == "new-host"

    @patch("smplkit.config.client.list_configs.asyncio_detailed")
    def test_refresh_fires_listeners(self, mock_list):
        client = _make_connected_async_client({"db": {"host": "old"}})

        events = []
        client.config.on_change(lambda e: events.append(e))

        mock_attrs = MagicMock()
        mock_attrs.name = "DB"
        mock_attrs.key = "db"
        mock_attrs.description = None
        mock_attrs.parent = None
        mock_attrs.items = None
        mock_attrs.environments = None
        mock_attrs.created_at = None
        mock_attrs.updated_at = None

        mock_resource = MagicMock()
        mock_resource.id = "cfg-1"
        mock_resource.attributes = mock_attrs

        mock_parsed = MagicMock()
        mock_parsed.data = [mock_resource]

        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.content = b""
        mock_response.parsed = mock_parsed

        async def fake_list(*args, **kwargs):
            return mock_response

        mock_list.side_effect = fake_list

        async def run():
            with patch(
                "smplkit.config.models.AsyncConfig._build_chain",
                return_value=[{"id": "cfg-1", "values": {"host": "new-host"}, "environments": {}}],
            ):
                await client.config.refresh()

        asyncio.run(run())
        assert len(events) == 1
        assert events[0].source == "manual"


class TestAsyncChangeListeners:
    def test_on_change_registers_listener(self):
        client = _make_connected_async_client()
        client.config.on_change(lambda e: None)
        assert len(client.config._listeners) == 1

    def test_on_change_with_filters(self):
        client = _make_connected_async_client()
        client.config.on_change(lambda e: None, config_key="db", item_key="host")
        assert client.config._listeners[0][1] == "db"
        assert client.config._listeners[0][2] == "host"

    def test_fire_change_listeners(self):
        client = _make_connected_async_client()
        events = []
        client.config.on_change(lambda e: events.append(e))
        client.config._fire_change_listeners(
            {"db": {"host": "old"}},
            {"db": {"host": "new"}},
            source="manual",
        )
        assert len(events) == 1
        assert events[0].config_key == "db"
        assert events[0].item_key == "host"

    def test_fire_change_listeners_filters_work(self):
        client = _make_connected_async_client()
        events = []
        client.config.on_change(lambda e: events.append(e), config_key="db", item_key="host")
        client.config._fire_change_listeners(
            {"db": {"host": "old", "port": 1}, "other": {"x": 1}},
            {"db": {"host": "new", "port": 2}, "other": {"x": 2}},
            source="manual",
        )
        assert len(events) == 1
        assert events[0].item_key == "host"

    def test_no_change_fires_nothing(self):
        client = _make_connected_async_client()
        events = []
        client.config.on_change(lambda e: events.append(e))
        client.config._fire_change_listeners(
            {"db": {"host": "same"}},
            {"db": {"host": "same"}},
            source="manual",
        )
        assert len(events) == 0

    def test_listener_exception_is_caught(self):
        client = _make_connected_async_client()
        good_events = []

        def bad_listener(event):
            raise ValueError("boom")

        client.config.on_change(bad_listener)
        client.config.on_change(lambda e: good_events.append(e))

        client.config._fire_change_listeners(
            {"db": {"host": "old"}},
            {"db": {"host": "new"}},
            source="manual",
        )
        assert len(good_events) == 1



"""Tests for auto-discovery of Python loggers."""

import logging as stdlib_logging

from smplkit.logging._discovery import (
    discover_existing_loggers,
    install_discovery_patch,
    uninstall_discovery_patch,
)


class TestDiscoverExistingLoggers:
    def test_finds_root(self):
        loggers = discover_existing_loggers()
        names = [name for name, _ in loggers]
        assert "root" in names

    def test_finds_pre_existing_logger(self):
        # Create a logger before discovery
        test_name = "test.discover.pre_existing_abc123"
        stdlib_logging.getLogger(test_name)
        loggers = discover_existing_loggers()
        names = [name for name, _ in loggers]
        assert test_name in names

    def test_returns_effective_level(self):
        test_name = "test.discover.level_check_xyz"
        lg = stdlib_logging.getLogger(test_name)
        lg.setLevel(stdlib_logging.ERROR)
        loggers = discover_existing_loggers()
        level_map = dict(loggers)
        assert level_map[test_name] == stdlib_logging.ERROR

    def test_skips_placeholder_objects(self):
        # PlaceHolders are created for intermediate names
        # e.g., getLogger("a.b.c") creates PlaceHolder for "a" and "a.b"
        # Verify we only get Logger instances
        test_name = "test.placeholder.skip.deep.logger"
        stdlib_logging.getLogger(test_name)
        loggers = discover_existing_loggers()
        for name, _ in loggers:
            obj = stdlib_logging.root.manager.loggerDict.get(name)
            if obj is not None:
                assert isinstance(obj, stdlib_logging.Logger)


class TestDiscoveryPatch:
    def setup_method(self):
        # Ensure patch is uninstalled before each test
        uninstall_discovery_patch()

    def teardown_method(self):
        uninstall_discovery_patch()

    def test_detects_new_logger(self):
        discovered = []

        def on_new(name, level):
            discovered.append((name, level))

        install_discovery_patch(on_new)
        test_name = "test.patch.new_logger_detect_001"
        stdlib_logging.getLogger(test_name)
        assert any(name == test_name for name, _ in discovered)

    def test_does_not_fire_for_existing_logger(self):
        test_name = "test.patch.existing_no_fire_002"
        stdlib_logging.getLogger(test_name)  # Create before patch

        discovered = []

        def on_new(name, level):
            discovered.append((name, level))

        install_discovery_patch(on_new)
        stdlib_logging.getLogger(test_name)  # Get again — not new
        assert not any(name == test_name for name, _ in discovered)

    def test_uninstall_restores_original(self):
        discovered = []

        def on_new(name, level):
            discovered.append((name, level))

        install_discovery_patch(on_new)
        uninstall_discovery_patch()

        test_name = "test.patch.after_uninstall_003"
        stdlib_logging.getLogger(test_name)
        assert not any(name == test_name for name, _ in discovered)

    def test_double_install_is_noop(self):
        calls_1 = []
        calls_2 = []

        install_discovery_patch(lambda n, _lv: calls_1.append(n))
        install_discovery_patch(lambda n, _lv: calls_2.append(n))

        test_name = "test.patch.double_install_004"
        stdlib_logging.getLogger(test_name)
        # First callback should be active
        assert test_name in calls_1
        # Second was ignored
        assert test_name not in calls_2



"""Tests for API key resolution chain."""

from __future__ import annotations

import pytest

from smplkit import AsyncSmplClient, SmplClient, SmplError
from smplkit._resolve import _resolve_api_key


class TestResolveApiKey:
    """Unit tests for the _resolve_api_key helper."""

    def test_explicit_key_returned(self):
        assert _resolve_api_key("sk_api_explicit") == "sk_api_explicit"

    def test_env_var_used_when_no_explicit(self, monkeypatch):
        monkeypatch.setenv("SMPLKIT_API_KEY", "sk_api_env")
        assert _resolve_api_key(None) == "sk_api_env"

    def test_config_file_used_when_no_explicit_no_env(self, monkeypatch, tmp_path):
        monkeypatch.delenv("SMPLKIT_API_KEY", raising=False)
        config_file = tmp_path / ".smplkit"
        config_file.write_text("[default]\napi_key = sk_api_file\n")
        monkeypatch.setattr("smplkit._resolve.Path.home", lambda: tmp_path)
        assert _resolve_api_key(None) == "sk_api_file"

    def test_none_when_no_key_anywhere(self, monkeypatch, tmp_path):
        monkeypatch.delenv("SMPLKIT_API_KEY", raising=False)
        monkeypatch.setattr("smplkit._resolve.Path.home", lambda: tmp_path)
        assert _resolve_api_key(None) is None

    def test_none_when_file_has_no_api_key(self, monkeypatch, tmp_path):
        monkeypatch.delenv("SMPLKIT_API_KEY", raising=False)
        config_file = tmp_path / ".smplkit"
        config_file.write_text("[default]\nother_key = value\n")
        monkeypatch.setattr("smplkit._resolve.Path.home", lambda: tmp_path)
        assert _resolve_api_key(None) is None

    def test_none_when_file_is_malformed(self, monkeypatch, tmp_path):
        monkeypatch.delenv("SMPLKIT_API_KEY", raising=False)
        config_file = tmp_path / ".smplkit"
        config_file.write_text("this is not valid ini {{{}}")
        monkeypatch.setattr("smplkit._resolve.Path.home", lambda: tmp_path)
        assert _resolve_api_key(None) is None

    def test_explicit_takes_precedence_over_env(self, monkeypatch):
        monkeypatch.setenv("SMPLKIT_API_KEY", "sk_api_env")
        assert _resolve_api_key("sk_api_explicit") == "sk_api_explicit"

    def test_env_takes_precedence_over_file(self, monkeypatch, tmp_path):
        monkeypatch.setenv("SMPLKIT_API_KEY", "sk_api_env")
        config_file = tmp_path / ".smplkit"
        config_file.write_text("[default]\napi_key = sk_api_file\n")
        monkeypatch.setattr("smplkit._resolve.Path.home", lambda: tmp_path)
        assert _resolve_api_key(None) == "sk_api_env"

    def test_empty_env_var_treated_as_unset(self, monkeypatch, tmp_path):
        monkeypatch.setenv("SMPLKIT_API_KEY", "")
        config_file = tmp_path / ".smplkit"
        config_file.write_text("[default]\napi_key = sk_api_file\n")
        monkeypatch.setattr("smplkit._resolve.Path.home", lambda: tmp_path)
        assert _resolve_api_key(None) == "sk_api_file"

    def test_comments_are_ignored(self, monkeypatch, tmp_path):
        monkeypatch.delenv("SMPLKIT_API_KEY", raising=False)
        config_file = tmp_path / ".smplkit"
        config_file.write_text("# comment\n[default]\n# another comment\napi_key = sk_api_comment\n")
        monkeypatch.setattr("smplkit._resolve.Path.home", lambda: tmp_path)
        assert _resolve_api_key(None) == "sk_api_comment"

    def test_missing_default_section(self, monkeypatch, tmp_path):
        monkeypatch.delenv("SMPLKIT_API_KEY", raising=False)
        config_file = tmp_path / ".smplkit"
        config_file.write_text("[staging]\napi_key = sk_api_staging\n")
        monkeypatch.setattr("smplkit._resolve.Path.home", lambda: tmp_path)
        assert _resolve_api_key(None) is None

    def test_default_section_without_api_key(self, monkeypatch, tmp_path):
        monkeypatch.delenv("SMPLKIT_API_KEY", raising=False)
        config_file = tmp_path / ".smplkit"
        config_file.write_text("[default]\nsome_other = value\n")
        monkeypatch.setattr("smplkit._resolve.Path.home", lambda: tmp_path)
        assert _resolve_api_key(None) is None


class TestSmplClientResolution:
    """Integration tests: SmplClient constructor uses resolution chain."""

    def test_explicit_key(self):
        client = SmplClient(api_key="sk_api_test", environment="test")
        assert client._api_key == "sk_api_test"

    def test_env_var(self, monkeypatch):
        monkeypatch.setenv("SMPLKIT_API_KEY", "sk_api_env")
        client = SmplClient(environment="test")
        assert client._api_key == "sk_api_env"

    def test_config_file(self, monkeypatch, tmp_path):
        monkeypatch.delenv("SMPLKIT_API_KEY", raising=False)
        config_file = tmp_path / ".smplkit"
        config_file.write_text("[default]\napi_key = sk_api_file\n")
        monkeypatch.setattr("smplkit._resolve.Path.home", lambda: tmp_path)
        client = SmplClient(environment="test")
        assert client._api_key == "sk_api_file"

    def test_error_when_no_key(self, monkeypatch, tmp_path):
        monkeypatch.delenv("SMPLKIT_API_KEY", raising=False)
        monkeypatch.setattr("smplkit._resolve.Path.home", lambda: tmp_path)
        with pytest.raises(SmplError, match="No API key provided"):
            SmplClient(environment="test")

    def test_error_message_lists_all_methods(self, monkeypatch, tmp_path):
        monkeypatch.delenv("SMPLKIT_API_KEY", raising=False)
        monkeypatch.setattr("smplkit._resolve.Path.home", lambda: tmp_path)
        with pytest.raises(SmplError, match="Pass api_key to the constructor"):
            SmplClient(environment="test")
        with pytest.raises(SmplError, match="SMPLKIT_API_KEY"):
            SmplClient(environment="test")
        with pytest.raises(SmplError, match=r"~/.smplkit"):
            SmplClient(environment="test")

    def test_error_when_no_environment(self):
        with pytest.raises(SmplError, match="No environment provided"):
            SmplClient(api_key="sk_api_test")

    def test_environment_from_env_var(self, monkeypatch):
        monkeypatch.setenv("SMPLKIT_ENVIRONMENT", "staging")
        client = SmplClient(api_key="sk_api_test")
        assert client._environment == "staging"

    def test_environment_explicit_takes_precedence(self, monkeypatch):
        monkeypatch.setenv("SMPLKIT_ENVIRONMENT", "staging")
        client = SmplClient(api_key="sk_api_test", environment="production")
        assert client._environment == "production"

    def test_service_from_env_var(self, monkeypatch):
        monkeypatch.setenv("SMPLKIT_SERVICE", "user-service")
        client = SmplClient(api_key="sk_api_test", environment="test")
        assert client._service == "user-service"

    def test_service_explicit_takes_precedence(self, monkeypatch):
        monkeypatch.setenv("SMPLKIT_SERVICE", "env-service")
        client = SmplClient(api_key="sk_api_test", environment="test", service="explicit-service")
        assert client._service == "explicit-service"

    def test_service_optional(self):
        client = SmplClient(api_key="sk_api_test", environment="test")
        assert client._service is None


class TestAsyncSmplClientResolution:
    """Integration tests: AsyncSmplClient constructor uses resolution chain."""

    def test_explicit_key(self):
        client = AsyncSmplClient(api_key="sk_api_test", environment="test")
        assert client._api_key == "sk_api_test"

    def test_env_var(self, monkeypatch):
        monkeypatch.setenv("SMPLKIT_API_KEY", "sk_api_env")
        client = AsyncSmplClient(environment="test")
        assert client._api_key == "sk_api_env"

    def test_error_when_no_key(self, monkeypatch, tmp_path):
        monkeypatch.delenv("SMPLKIT_API_KEY", raising=False)
        monkeypatch.setattr("smplkit._resolve.Path.home", lambda: tmp_path)
        with pytest.raises(SmplError, match="No API key provided"):
            AsyncSmplClient(environment="test")

    def test_error_when_no_environment(self):
        with pytest.raises(SmplError, match="No environment provided"):
            AsyncSmplClient(api_key="sk_api_test")

    def test_environment_from_env_var(self, monkeypatch):
        monkeypatch.setenv("SMPLKIT_ENVIRONMENT", "staging")
        client = AsyncSmplClient(api_key="sk_api_test")
        assert client._environment == "staging"

    def test_service_optional(self):
        client = AsyncSmplClient(api_key="sk_api_test", environment="test")
        assert client._service is None
